@@ -196,6 +196,7 @@ static i_fill_solid_t base_solid_fill_comb =
 =item i_fill_destroy(fill)
 
 =category Fills
+=synopsis i_fill_destroy(fill);
 
 Call to destroy any fill object.
 
@@ -213,6 +214,7 @@ i_fill_destroy(i_fill_t *fill) {
 =item i_new_fill_solidf(color, combine)
 
 =category Fills
+=synopsis i_fill_t *fill = i_new_fill_solidf(&fcolor, combine);
 
 Create a solid fill based on a float color.
 
@@ -244,6 +246,7 @@ i_new_fill_solidf(const i_fcolor *c, int combine) {
 =item i_new_fill_solid(color, combine)
 
 =category Fills
+=synopsis i_fill_t *fill = i_new_fill_solid(&color, combine);
 
 Create a solid fill based on an 8-bit color.
 
@@ -427,6 +430,7 @@ i_new_hatch_low(const i_color *fg, const i_color *bg, const i_fcolor *ffg, const
 =item i_new_fill_hatch(fg, bg, combine, hatch, cust_hatch, dx, dy)
 
 =category Fills
+=synopsis i_fill_t *fill = i_new_fill_hatch(&fg_color, &bg_color, combine, hatch, custom_hatch, dx, dy);
 
 Creates a new hatched fill with the fg color used for the 1 bits in
 the hatch and bg for the 0 bits.  If combine is non-zero alpha values
@@ -452,6 +456,7 @@ i_new_fill_hatch(const i_color *fg, const i_color *bg, int combine, int hatch,
 =item i_new_fill_hatchf(fg, bg, combine, hatch, cust_hatch, dx, dy)
 
 =category Fills
+=synopsis i_fill_t *fill = i_new_fill_hatchf(&fg_fcolor, &bg_fcolor, combine, hatch, custom_hatch, dx, dy);
 
 Creates a new hatched fill with the fg color used for the 1 bits in
 the hatch and bg for the 0 bits.  If combine is non-zero alpha values
@@ -489,6 +494,7 @@ struct i_fill_image_t {
 =item i_new_fill_image(im, matrix, xoff, yoff, combine)
 
 =category Fills
+=synopsis i_fill_t *fill = i_new_fill_image(src_img, matrix, x_offset, y_offset, combine);
 
 Create an image based fill.
 
@@ -621,10 +627,25 @@ i_new_hatch_low(const i_color *fg, const i_color *bg,
   fill->base.fill_with_color = fill_hatch;
   fill->base.fill_with_fcolor = fill_hatchf;
   fill->base.destroy = NULL;
-  fill->fg = fg ? *fg : fcolor_to_color(ffg);
-  fill->bg = bg ? *bg : fcolor_to_color(fbg);
-  fill->ffg = ffg ? *ffg : color_to_fcolor(fg);
-  fill->fbg = fbg ? *fbg : color_to_fcolor(bg);
+  /* Some Sun C didn't like the condition expressions that were here.
+     See https://rt.cpan.org/Ticket/Display.html?id=21944
+   */
+  if (fg)
+    fill->fg = *fg;
+  else
+    fill->fg = fcolor_to_color(ffg);
+  if (bg)
+    fill->bg = *bg;
+  else
+    fill->bg = fcolor_to_color(fbg);
+  if (ffg) 
+    fill->ffg = *ffg;
+  else
+    fill->ffg = color_to_fcolor(fg);
+  if (fbg)
+    fill->fbg = *fbg;
+  else
+    fill->fbg = color_to_fcolor(bg);
   if (combine) {
     i_get_combine(combine, &fill->base.combine, &fill->base.combinef);
   }
@@ -661,7 +682,10 @@ static void fill_hatch(i_fill_t *fill, int x, int y, int width, int channels,
   int mask = 128 >> xpos;
 
   while (width-- > 0) {
-    *data++ = (byte & mask) ? f->fg : f->bg;
+    if (byte & mask)
+      *data++ = f->fg;
+    else
+      *data++ = f->bg;
     
     if ((mask >>= 1) == 0)
       mask = 128;
@@ -683,7 +707,10 @@ static void fill_hatchf(i_fill_t *fill, int x, int y, int width, int channels,
   int mask = 128 >> xpos;
   
   while (width-- > 0) {
-    *data++ = (byte & mask) ? f->ffg : f->fbg;
+    if (byte & mask)
+      *data++ = f->ffg;
+    else
+      *data++ = f->fbg;
     
     if ((mask >>= 1) == 0)
       mask = 128;
