@@ -2,6 +2,9 @@ package Imager::Font;
 
 use Imager::Color;
 use strict;
+use vars qw($VERSION);
+
+$VERSION = "1.033";
 
 # the aim here is that we can:
 #  - add file based types in one place: here
@@ -14,16 +17,19 @@ my %drivers =
         class=>'Imager::Font::Truetype',
         module=>'Imager/Font/Truetype.pm',
         files=>'.*\.ttf$',
+	description => 'FreeType 1.x',
        },
    t1=>{
         class=>'Imager::Font::Type1',
         module=>'Imager/Font/Type1.pm',
         files=>'.*\.pfb$',
+	description => 'T1Lib',
        },
    ft2=>{
          class=>'Imager::Font::FreeType2',
          module=>'Imager/Font/FreeType2.pm',
-         files=>'.*\.(pfa|pfb|otf|ttf|fon|fnt)$',
+         files=>'.*\.(pfa|pfb|otf|ttf|fon|fnt|dfont|pcf(\.gz)?)$',
+	 description => 'FreeType 2.x',
         },
    ifs=>{
          class=>'Imager::Font::Image',
@@ -33,6 +39,7 @@ my %drivers =
    w32=>{
          class=>'Imager::Font::Win32',
          module=>'Imager/Font/Win32.pm',
+	 description => 'Win32 GDI Fonts',
         },
   );
 
@@ -43,7 +50,9 @@ my @priority = qw(t1 tt ft2 ifs);
 # this function is called from Imager.pm to finish initialization
 sub __init {
   @priority = grep Imager::i_has_format($_), @priority;
-  delete @drivers{grep !Imager::i_has_format($_), keys %drivers};
+  for my $driver_name (grep Imager::i_has_format($_), keys %drivers) {
+    $drivers{$driver_name}{enabled} = 1;
+  }
 }
 
 # search method
@@ -73,7 +82,7 @@ sub new {
     }
 
     $type = $hsh{'type'};
-    if (!defined($type) or !$drivers{$type}) {
+    if (!defined($type) or !$drivers{$type} or !$drivers{$type}{enabled}) {
       for my $drv (@priority) {
         undef $type;
         my $re = $drivers{$drv}{files} or next;
@@ -84,7 +93,21 @@ sub new {
       }
     }
     if (!defined($type)) {
-      $Imager::ERRSTR = "Font type not found";
+      # some types we can support, but the driver isn't available
+      # work out which drivers support it, so we can provide the user
+      # some useful information on how to get it working
+      my @not_here;
+      for my $driver_name (keys %drivers) {
+	my $driver = $drivers{$driver_name};
+	push @not_here, "$driver_name ($driver->{description})"
+	  if $driver->{files} && $file =~ /$driver->{files}/i;
+      }
+      if (@not_here) {
+	$Imager::ERRSTR = "No font drivers enabled that can support this file, rebuild Imager with any of ".join(", ", @not_here)." to use this font file";
+      }
+      else {
+	$Imager::ERRSTR = "No font type found for $hsh{'file'}";
+      }
       return;
     }
   } elsif ($hsh{face}) {
@@ -120,9 +143,10 @@ sub draw {
     $Imager::ERRSTR = 'No image supplied to $font->draw()';
     return;
   }
+  my $image = $input{image};
   $input{string} = _first($input{string}, $input{text});
   unless (defined $input{string}) {
-    $Imager::ERRSTR = "Missing required parameter 'string'";
+    $image->_set_error("Missing required parameter 'string'");
     return;
   }
   $input{aa} = _first($input{aa}, $input{antialias}, $self->{aa}, 1);
@@ -133,14 +157,19 @@ sub draw {
 
   $input{size} = _first($input{size}, $self->{size});
   unless (defined $input{size}) {
-    $input{image}{ERRSTR} = "No font size provided";
+    $image->_set_error("No font size provided");
     return undef;
   }
   $input{align} = _first($input{align}, 1);
   $input{utf8} = _first($input{utf8}, $self->{utf8}, 0);
   $input{vlayout} = _first($input{vlayout}, $self->{vlayout}, 0);
 
-  $self->_draw(%input);
+  my $result = $self->_draw(%input);
+  unless ($result) {
+    $image->_set_error($image->_error_as_msg());
+  }
+
+  return $result;
 }
 
 sub align {
@@ -148,17 +177,20 @@ sub align {
   my %input = ( halign => 'left', valign => 'baseline', 
                 'x' => 0, 'y' => 0, @_ );
 
-  my $text = _first($input{string}, $input{text});
-  unless (defined $text) {
-    Imager->_set_error("Missing required parameter 'string'");
-    return;
-  }
-
   # image needs to be supplied, but can be supplied as undef
   unless (exists $input{image}) {
     Imager->_set_error("Missing required parameter 'image'");
     return;
   }
+
+  my $errors_to = $input{image} || 'Imager';
+
+  my $text = _first($input{string}, $input{text});
+  unless (defined $text) {
+    $errors_to->_set_error("Missing required parameter 'string'");
+    return;
+  }
+
   my $size = _first($input{size}, $self->{size});
   my $utf8 = _first($input{utf8}, 0);
 
@@ -194,8 +226,11 @@ sub align {
   elsif ($halign eq 'center') {
     $x -= $bbox->start_offset + $bbox->total_width / 2;
   }
-  elsif ($halign eq 'end' || $halign eq 'right') {
-    $x -= $bbox->start_offset + $bbox->total_width - 1;
+  elsif ($halign eq 'end') {
+    $x -= $bbox->advance_width;
+  }
+  elsif ($halign eq 'right') {
+    $x -= $bbox->advance_width - $bbox->right_bearing;
   }
   $x = int($x);
   $y = int($y);
@@ -204,14 +239,6 @@ sub align {
     delete @input{qw/x y/};
     $self->draw(%input, 'x' => $x, 'y' => $y, align=>1)
       or return;
-#      for my $i (1 .. length $text) {
-#        my $work = substr($text, 0, $i);
-#        my $bbox = $self->bounding_box(string=>$work, size=>$size, utf8=>$utf8);
-#        my $nx = $x + $bbox->end_offset;
-#        $input{image}->setpixel(x=>[ ($nx) x 5 ],
-#                                'y'=>[ $y-2, $y-1, $y, $y+1, $y+2 ],
-#                                color=>'FF0000');
-#      }
   }
 
   return ($x+$bbox->start_offset, $y-$bbox->ascent, 
@@ -325,15 +352,13 @@ Imager::Font - Font handling for Imager.
    $pos_width,
    $global_ascent,
    $descent,
-   $ascent) = $font->bounding_box(string=>"Foo");
-
-  $logo = $font->logo(text   => "Slartibartfast Enterprises",
-		      size   => 40,
-		      border => 5,
-		      color  => $green);
-  # logo is proposed - doesn't exist yet
+   $ascent,
+   $advance_width,
+   $right_bearing) = $font->bounding_box(string=>"Foo");
 
+  my $bbox_object = $font->bounding_box(string=>"Foo");
 
+  # documented in Imager::Draw
   $img->string(font  => $font,
 	     text  => "Model-XYZ",
 	     x     => 15,
@@ -342,8 +367,6 @@ Imager::Font - Font handling for Imager.
 	     color => $red,
 	     aa    => 1);
 
-  # Documentation in Imager.pm
-
 =head1 DESCRIPTION
 
 This module handles creating Font objects used by imager.  The module
@@ -366,6 +389,7 @@ this:
 This creates a font object to pass to functions that take a font argument.
 
   $font = Imager::Font->new(file  => 'denmark.ttf',
+                            index => 0,
 			    color => $blue,
 			    size  => 30,
 			    aa    => 1);
@@ -381,13 +405,17 @@ you can tell it explicitly by using the C<type> parameter:
   $t1font = Imager::Font->new(file => 'fruitcase', type => 't1');
   $ttfont = Imager::Font->new(file => 'arglebarf', type => 'tt');
 
+The C<index> parameter is used to select a single face from a font
+file containing more than one face, for example, from a Macintosh font
+suitcase or a .dfont file.
+
 If any of the C<color>, C<size> or C<aa> parameters are omitted when
 calling C<Imager::Font->new()> the they take the following values:
 
-
   color => Imager::Color->new(255, 0, 0, 0);  # this default should be changed
   size  => 15
   aa    => 0
+  index => 0
 
 To use Win32 fonts supply the facename of the font:
 
@@ -399,6 +427,85 @@ contruct the full name of the font as above.
 
 Other logical font attributes may be added if there is sufficient demand.
 
+Parameters:
+
+=over
+
+=item *
+
+file - name of the file to load the font from.
+
+=item *
+
+face - face name.  This is used only under Win32 to create a GDI based
+font.  This is ignored if the C<file> parameter is supplied.
+
+=item *
+
+type - font driver to use.  Currently the permitted values for this are:
+
+=over
+
+=item *
+
+tt - Freetype 1.x driver.  Supports TTF fonts.
+
+=item *
+
+t1 - T1 Lib driver.  Supports Postscript Type 1 fonts.  Allows for
+synthesis of underline, strikethrough and overline.
+
+=item *
+
+ft2 - Freetype 2.x driver.  Supports many different font formats.
+Also supports the transform() method.
+
+=back
+
+=item *
+
+color - the default color used with this font.  Default: red.
+
+=item *
+
+size - the default size used with this font.  Default: 15.
+
+=item *
+
+utf8 - if non-zero then text supplied to $img->string(...) and
+$font->bounding_box(...) is assumed to be UTF 8 encoded by default.
+
+=item *
+
+align - the default value for the $img->string(...) C<align>
+parameter.  Default: 1.
+
+=item *
+
+vlayout - the default value for the $img->string(...) C<vlayout>
+parameter.  Default: 0.
+
+=item *
+
+aa - the default value for the $im->string(...) C<aa> parameter.
+Default: 0.
+
+=item *
+
+index - for font file containing multiple fonts this selects which
+font to use.  This is useful for Macintosh DFON (.dfont) and suitcase
+font files.
+
+If you want to use a suitcase font you will need to tell Imager to use
+the FreeType 2.x driver by setting C<type> to C<'ft2'>:
+
+  my $font = Imager::Font->new(file=>$file, index => 1, type=>'ft2')
+    or die Imager->errstr;
+
+=back
+
+
+
 =item bounding_box
 
 Returns the bounding box for the specified string.  Example:
@@ -409,7 +516,8 @@ Returns the bounding box for the specified string.  Example:
       $global_ascent,
       $descent,
       $ascent,
-      $advance_width) = $font->bounding_box(string => "A Fool");
+      $advance_width,
+      $right_bearing) = $font->bounding_box(string => "A Fool");
 
   my $bbox_object = $font->bounding_box(string => "A Fool");
 
@@ -447,6 +555,12 @@ start at, this is often the same as C<$pos_width>, but can be
 different if the final character overlaps the right side of its
 character cell.
 
+=item C<$right_bearing>
+
+The distance from the right side of the final glyph to the end of the
+advance width.  If the final glyph overflows the advance width this
+value is negative.
+
 =back
 
 Obviously we can stuff all the results into an array just as well:
@@ -475,94 +589,49 @@ Returns an L<Imager::Font::BBox> object in scalar context, so you can
 avoid all those confusing indices.  This has methods as named above,
 with some extra convenience methods.
 
-=item string
-
-This is a method of the Imager class but because it's described in
-here since it belongs to the font routines.  Example:
-
-  $img=Imager->new();
-  $img->read(file=>"test.jpg");
-  $img->string(font=>$t1font,
-               text=>"Model-XYZ",
-               x=>0,
-               y=>40,
-               size=>40,
-               color=>$red);
-  $img->write(file=>"testout.jpg");
-
-This would put a 40 pixel high text in the top left corner of an
-image.  If you measure the actuall pixels it varies since the fonts
-usually do not use their full height.  It seems that the color and
-size can be specified twice.  When a font is created only the actual
-font specified matters.  It his however convenient to store default
-values in a font, such as color and size.  If parameters are passed to
-the string function they are used instead of the defaults stored in
-the font.
-
-The following parameters can be supplied to the string() method:
+Parameters are:
 
 =over
 
-=item string
+=item *
 
-The text to be rendered.  If this isn't present the 'text' parameter
-is used.  If neither is present the call will fail.
+string - the string to calculate the bounding box for.  Required.
 
-=item aa
+=item *
 
-If non-zero the output will be anti-aliased.
+size - the font size to use.  Default: value set in
+Imager::Font->new(), or 15.
 
-=item x
+=item *
 
-=item y
+sizew - the font width to use.  Default to the value of the C<size>
+parameter.
 
-The start point for rendering the text.  See the align parameter.
+=item *
 
-=item align
+utf8 - For drivers that support it, treat the string as UTF8 encoded.
+For versions of perl that support Unicode (5.6 and later), this will
+be enabled automatically if the 'string' parameter is already a UTF8
+string. See L<UTF8> for more information.  Default: the C<utf8> value
+passed to Imager::Font->new(...) or 0.
 
-If non-zero the point supplied in (x,y) will be on the base-line, if
-zero then (x,y) will be at the top-left of the first character.
+=item *
 
-=item channel
+x, y - offsets applied to @box[0..3] to give you a adjusted bounding
+box.  Ignored in scalar context.
 
-If present, the text will be written to the specified channel of the
-image and the color parameter will be ignore.
+=item *
 
-=item color
-
-The color to draw the text in.
-
-=item size
-
-The point-size to draw the text at.
-
-=item sizew
-
-For drivers that support it, the width to draw the text at.  Defaults
-to be value of the 'size' parameter.
-
-=item utf8
-
-For drivers that support it, treat the string as UTF8 encoded.  For
-versions of perl that support Unicode (5.6 and later), this will be
-enabled automatically if the 'string' parameter is already a UTF8
-string. See L<UTF8> for more information.
-
-=item vlayout
-
-For drivers that support it, draw the text vertically.  Note: I
-haven't found a font that has the appropriate metrics yet.
+canon - if non-zero and the C<x>, C<y> parameters are not supplied,
+then $pos_width and $global_ascent values will returned as the width
+and height of the text instead.
 
 =back
 
-If string() is called with the C<channel> parameter then the color
-isn't used and the font is drawn in only one channel of the image.
-This can be quite handy to create overlays.  See the examples for tips
-about this.
+=item string
 
-Sometimes it is necessary to know how much space a string takes before
-rendering it.  The bounding_box() method described earlier can be used
-for that.
+The $img->string(...) method is now documented in
+L<Imager::Draw/string>
 
 =item align(string=>$text, size=>$size, x=>..., y=>..., valign => ..., halign=>...)
 
@@ -570,7 +639,7 @@ Higher level text output - outputs the text aligned as specified
 around the given point (x,y).
 
   # "Hello" centered at 100, 100 in the image.
-  my ($left, $top, $bottom, $right) = 
+  my ($left, $top, $right, $bottom) = 
     $font->align(string=>"Hello",
                  x=>100, y=>100, 
                  halign=>'center', valign=>'center', 
@@ -627,9 +696,7 @@ The point is at the right end of the text.
 
 =item end
 
-The point is at the right end of the text.  This will change to the
-end point of the text (once the proper bounding box interfaces are
-available).
+The point is at the end point of the text.
 
 =back
 
@@ -648,11 +715,28 @@ Returns a list specifying the bounds of the drawn text.
 
 =item dpi(dpi=>$dpi)
 
-Set retrieve the spatial resolution of the image in dots per inch.
+Set or retrieve the spatial resolution of the image in dots per inch.
 The default is 72 dpi.
 
 This isn't implemented for all font types yet.
 
+Possible parameters are:
+
+=over
+
+=item *
+
+xdpi, ydpi - set the horizontal and vertical resolution in dots per
+inch.
+
+=item *
+
+dpi - set both horizontal and vertical resolution to this value.
+
+=back
+
+Returns a list containing the previous xdpi, ydpi values.
+
 =item transform(matrix=>$matrix)
 
 Applies a transformation to the font, where matrix is an array ref of
@@ -670,6 +754,12 @@ See the end of the test script t/t38ft2font.t for an example.
 Currently only the ft2 (Freetype 2.x) driver supports the transform()
 method.
 
+See samples/slant_text.pl for a sample using this function.
+
+Note that the transformation is done in font co-ordinates where y
+increases as you move up, not image co-ordinates where y decreases as
+you move up.
+
 =item has_chars(string=>$text)
 
 Checks if the characters in $text are defined by the font.
@@ -682,18 +772,22 @@ characters.  Supports UTF8 where the font driver supports UTF8.
 Not all fonts support this method (use $font->can("has_chars") to
 check.)
 
-=item logo
+=over
 
-This method doesn't exist yet but is under consideration.  It would mostly
-be helpful for generating small tests and such.  Its proposed interface is:
+=item *
 
-  $img = $font->logo(string=>"Plan XYZ", color=>$blue, border=>7);
+string - string of characters to check for.  Required.  Must contain
+at least one character.
 
-This would be nice for writing (admittedly multiline) one liners like:
+=item *
 
-Imager::Font->new(file=>"arial.ttf", color=>$blue, aa=>1)
-            ->string(text=>"Plan XYZ", border=>5)
-            ->write(file=>"xyz.png");
+utf8 - For drivers that support it, treat the string as UTF8 encoded.
+For versions of perl that support Unicode (5.6 and later), this will
+be enabled automatically if the 'string' parameter is already a UTF8
+string. See L<UTF8> for more information.  Default: the C<utf8> value
+passed to Imager::Font->new(...) or 0.
+
+=back
 
 =item face_name()
 
@@ -720,8 +814,61 @@ say how those name tables are unreliable, or how FT2 handles them.
 Both Freetype 1.x and 2.x allow support for glyph names to not be
 included.
 
+=item draw
+
+This is used by Imager's string() method to implement drawing text.
+See L<Imager::Draw/string>.
+
 =back
 
+=head1 MULTIPLE MASTER FONTS
+
+The Freetype 2 driver supports multiple master fonts:
+
+=over
+
+=item is_mm()
+
+Test if the font is a multiple master font.
+
+=item mm_axes()
+
+Returns a list of the axes that can be changes in the font.  Each
+entry is an array reference which contains:
+
+=over
+
+=item 1.
+
+Name of the axis.
+
+=item 2.
+
+minimum value for this axis.
+
+=item 3.
+
+maximum value for this axis
+
+=back
+
+=item set_mm_coords(coords=>\@values)
+
+Blends an interpolated design from the master fonts.  @values must
+contain as many values as there are axes in the font.
+
+=back
+
+For example, to select the minimum value in each axis:
+
+  my @axes = $font->mm_axes;
+  my @coords = map $_->[1], @axes;
+  $font->set_mm_coords(coords=>\@coords);
+
+It's possible other drivers will support multiple master fonts in the
+future, check if your selected font object supports the is_mm() method
+using the can() method.
+
 =head1 UTF8
 
 There are 2 ways of rendering Unicode characters with Imager:
@@ -745,6 +892,16 @@ unicode character 00C3 "LATIN CAPITAL LETTER A WITH DIAERESIS", and
 your font doesn't support it, Imager will I<not> build it from 0041
 "LATIN CAPITAL LETTER A" and 0308 "COMBINING DIAERESIS".
 
+To check if a driver supports UTF8 call the utf8 method:
+
+=over
+
+=item utf8
+
+Return true if the font supports UTF8.
+
+=back
+
 =head2 Native UTF8 Support
 
 If your version of perl supports UTF8 and the driver supports UTF8,
@@ -783,7 +940,11 @@ Since some formats can be handled by more than one driver, a priority
 list is used to choose which one should be used, if a given format can
 be handled by more than one driver.
 
-The current priority can be retrieved with:
+=over
+
+=item priorities
+
+The current priorities can be retrieved with:
 
   @drivers = Imager::Font->priorities();
 
@@ -791,6 +952,8 @@ You can set new priorities and save the old priorities with:
 
   @old = Imager::Font->priorities(@drivers);
 
+=back
+
 If you supply driver names that are not currently supported, they will
 be ignored.
 
@@ -811,6 +974,19 @@ list.
 
 You need to modify this class to add new font types.
 
+The $pos_width member returned by the bounding_box() method has
+historically returned different values from different drivers.  The
+Freetype 1.x and 2.x, and the Win32 drivers return the max of the
+advance width and the right edge of the right-most glyph.  The Type 1
+driver always returns the right edge of the right-most glyph.
+
+The newer advance_width and right_bearing values allow access to any
+of the above.
+
+=head1 REVISION
+
+$Revision: 1263 $
+
 =head1 SEE ALSO
 
 Imager(3), Imager::Font::FreeType2(3), Imager::Font::Type1(3),
